import pytest
import jax
import jax.numpy as jnp
from jax import config
config.update("jax_enable_x64", True)
import numpy as np

from scipy.special import spherical_jn

from .functions import create_j_l, _j_0, _j_1

MAX_TEST_ORDER = 40
TEST_ORDERS = [i for i in range(MAX_TEST_ORDER)]
MAX_RADIUS = 50.0

r = jnp.linspace(1e-3, MAX_RADIUS, num=1001)
z = jnp.mgrid[1e-3:MAX_RADIUS:0.1, 1e-3:MAX_RADIUS:0.1]
z = -z[0] + 1j * z[1]

# Test _j_0 function
def test_j_0():
    expected = jnp.sinc(r / jnp.pi)
    assert np.allclose(_j_0(r), expected)

# Test _j_1 function
def test_j_1():
    expected = (jnp.sinc(r / jnp.pi) - jnp.cos(r)) / r
    assert np.allclose(_j_1(r), expected)

# Test invalid order raises ValueError
def test_create_j_l_invalid_order():
    with pytest.raises(ValueError):
        create_j_l(-1)


## Test for real valued arguments

@pytest.mark.parametrize("order", TEST_ORDERS)
def test_real_create_j_l_single_order(order):
    func = create_j_l(order=order, dtype=jnp.float64, output_all=False)
    expected = spherical_jn(order, r)
    assert np.allclose(func(r), expected, atol=1e-8, rtol=1e-8)


def test_real_create_j_l_output_all():
    func = create_j_l(order=MAX_TEST_ORDER, dtype=jnp.float64, output_all=True)
    expected = []
    for order in range(MAX_TEST_ORDER+1):
        expected.append(spherical_jn(order, r))
    expected = np.array(expected).swapaxes(0, 1)
    result = func(r)
    assert np.allclose(func(r), expected, atol=1e-8, rtol=1e-8)

@pytest.mark.parametrize("order", TEST_ORDERS)
def test_real_create_j_l_grad(order):
    func = create_j_l(order=order, dtype=jnp.float64, output_all=False)
    derivative_func = jnp.vectorize(jax.grad(func))
    expected = spherical_jn(order, r, derivative=True)
    assert np.allclose(derivative_func(r), expected, atol=1e-8, rtol=1e-8)


## Test for complex valued arguments

@pytest.mark.parametrize("order", TEST_ORDERS)
def test_complex_create_j_l_single_order(order):
    func = create_j_l(order=order, dtype=jnp.complex128, output_all=False)
    expected = spherical_jn(order, z)
    assert np.allclose(func(z), expected, rtol=1e-8)


def test_complex_create_j_l_output_all():
    func = create_j_l(order=MAX_TEST_ORDER, dtype=jnp.complex128, output_all=True)
    expected = []
    for order in range(MAX_TEST_ORDER+1):
        expected.append(spherical_jn(order, z))
    expected = np.array(expected).swapaxes(0, 1)
    result = func(z)
    assert np.allclose(func(z), expected, rtol=1e-8)


@pytest.mark.parametrize("order", TEST_ORDERS)
def test_complex_create_j_l_grad(order):
    func = create_j_l(order=order, dtype=jnp.complex128, output_all=False)
    derivative_func = jnp.vectorize(jax.grad(func, holomorphic=True))
    expected = spherical_jn(order, z, derivative=True)
    assert np.allclose(derivative_func(z), expected, rtol=1e-8)



#!/usr/bin/env python
# Copyright 2019-2022 The NeuralIL contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Code used to calculate the roots of the spherical Bessel functions of the
# first kind. Inspired on the C implementation by Jeremy K. Mason, also under
# an Apache license and included as part of the reference implementation of
# the spherical Bessel descriptors (doi: 10.1063/1.5111045 ,
# https://github.com/harharkh/sb_desc ). However, here we use the scipy root
# finding routines.

import jax
import numpy as np
import scipy as sp
import scipy.optimize

from . import functions

# The calculation starts with the roots of the order-0 function and
# proceeds upwards. Therefore, caching is essential.


class SphericalBesselRoots:
    """Calculator and cache of the roots of spherical Bessel functions.

    Access to the roots is supposed to happen directly through the "table"
    attribute. The first index is the order of the function, and the second
    is the order of the root. A negative value means that the corresponding
    root has not been calculated.

    Args:
        n_max: The maximum value of n to be used when generating the spherical
            Bessel descriptors. This affects which roots are calculated.
        dtype: The floating-point type to be used for the calculations.

    Raises:
        ValueError if n_max is negative.
    """
    def __init__(self, n_max: int, dtype: np.dtype = np.float32):
        if n_max < 0:
            raise ValueError("n_max cannot be negative")
        self.n_max = n_max
        # Initialize the table of precomputed values. Note that, even though
        # a rectangular NumPy array is used for convenience and performance,
        # the higher the order of the function, the fewer roots are computed.
        self.table = -np.ones((self.n_max + 1, self.n_max + 2))
        # The zeros of j_0(r) are trivially generated.
        self.table[0, :] = np.pi * np.arange(1, n_max + 3)
        # Proceed upwards from there, by using the fact that the roots of the
        # order-l function bracket the roots of the order-(l+1) function, per
        # Abramowitz and Stegun.
        for order in range(1, self.n_max + 1):
            # This could be made faster by using the SciPy implementation of
            # the spherical Bessels functions directly, but it only needs
            # to be done once and we get values of the roots that are
            # better adapted to our implementation.
            function = jax.jit(functions.create_j_l(order, dtype))
            for n in range(self.n_max + 2 - order):
                left = self.table[order - 1, n]
                right = self.table[order - 1, n + 1]
                self.table[order, n] = sp.optimize.brentq(function, left, right)


if __name__ == "__main__":
    # Time the calculation and check the maximum error in the roots.
    import datetime
    import timeit
    test_n_max = 140

    print(f"Building the table of roots up to n_max={test_n_max}")
    start = timeit.default_timer()
    roots = SphericalBesselRoots(test_n_max)
    end = timeit.default_timer()
    print("Time elapsed:", str(datetime.timedelta(seconds=end - start)))

    print("Computing the maximum error of the roots")
    start = timeit.default_timer()
    max_error = 0.
    for order in range(0, test_n_max + 1):
        function = functions.create_j_l(order)
        for n in range(test_n_max + 2 - order):
            max_error = max(max_error, function(roots.table[order, n]))
    end = timeit.default_timer()
    print("Time elapsed:", str(datetime.timedelta(seconds=end - start)))
    print("Maximum error:", max_error)



#!/usr/bin/env python
# Copyright 2019-2024 The NeuralIL and spbessax contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial

import jax
import jax.numpy as jnp
import numpy as onp


def _j_0(r: jnp.ndarray) -> jnp.ndarray:
    """Order-0 spherical Bessel function of the first kind.

    The evaluation is delegated to jax.numpy.sinc. The condition r >= 0 is
    not assessed.

    Args:
        r: The radial coordinate.

    Returns:
        The values of the order-0 spherical Bessel function of the first kind.
    """
    return jnp.sinc(r / jnp.pi)


def _j_1(r: jnp.ndarray) -> jnp.ndarray:
    """Order-1 spherical Bessel function of the first kind.

    The function is built on the basis of calls to existing jax.numpy code. The
    condition r >= 0 is not assessed.

    Args:
        r: The radial coordinate.

    Returns:
        The values of the order-1 spherical Bessel function of the first kind.
    """
    return (jnp.sinc(r / jnp.pi) - jnp.cos(r)) / r


def _calc_starting_order_Cai(order: int, r):
    """Use Cai's recipe to compute a starting point for his recurrence.

    When the method of Computer Physics Communications 182 (2011) 664,
    doi:10.1016/j.cpc.2010.11.019 is used to calculate j_l, an starting
    order is required for the descending recurrence. The original
    prescription works for any complex number, but this implementation
    only works for real arguments.

    Args:
        order: The order of the spherical Bessel function to be evaluated.
        r: A single value of the argument to the spherical bessel function.
    """
    cais_m = int(onp.floor(1.83 * r**0.91 + 9.))
    lower_bound = max(order + 1, cais_m)
    # This bound should depend on the kind of floating point used, but that
    # is not a critical point for our narrow use case.
    upper_bound = int(onp.floor(235. + 50. * onp.sqrt(r)))
    return min(lower_bound, upper_bound)


def create_j_l(order: int,
               output_all: bool = False,
               dtype: onp.dtype = onp.float32):
    """Generate the l-th spherical Bessel function of the first kind.

    This implementation is only intended for real arguments, and the focus is
    on compatibility with JAX. The calculation is handled by two different
    recurrence strategies, depending on whether the argument is less than or
    greater than the order.

    Args:
        order: The order of the spherical Bessel function, which cannot be
            negative.
        output_all: Whether to return all the values of the function up to the
            selected order or just the last one.
        dtype: A floating-point data type on which the function is intended to
           operate.

    Returns:
        A callable that takes radial coordinates and returns the values of
            the selected Bessel function.

    Raises:
        ValueError: if "order" is negative.
    """
    if order < 0:
        raise ValueError("the order of the function cannot be negative")
    elif order == 0:
        return _j_0
    elif order == 1:
        return ((lambda r: jnp.asarray([_j_0(r), _j_1(r)], dtype=dtype))
                if output_all else _j_1)

    def j_l_upward(r: jnp.ndarray) -> jnp.ndarray:
        """Order-l spherical Bessel function of the first kind with derivative.

        This implementation, intended for r>l, is based on ascendent recurrence
        and adapted from SciPy itself. The condition r >= 0 is not assessed.

        Args:
            r: The radial coordinate.

        Returns:
            The values of the spherical Bessel function of the first
            kind and the values of its derivative up to order l, in a tuple.
        """
        orders, derivatives = [], []
        orders.append(_j_0(r))
        orders.append(_j_1(r))
        derivatives.append(
            jnp.vectorize(jax.grad(_j_0, holomorphic=jnp.iscomplexobj(dtype)))(r))
        derivatives.append(
            jnp.vectorize(jax.grad(_j_1, holomorphic=jnp.iscomplexobj(dtype)))(r))

        # TODO: replace with jax.lax.scan operation
        for i in range(order - 1):
            temp = (2 * i + 3) / r * orders[-1] - orders[-2]
            orders.append(temp)
            derivatives.append(orders[-2] - (i + 3) * orders[-1] / r)

        orders = jnp.asarray(orders)
        derivatives = jnp.asarray(derivatives)

        return (orders, derivatives)

    starting_order = _calc_starting_order_Cai(order, order)
    # Cai's original prescription is 1e-305, which works for double
    # precision but causes huge problems with the derivative. eps**2 * r
    # seems to work well, since smaller r require a smaller value.
    initial_prefactor = onp.finfo(dtype).eps**2

    def j_l_Cai(r: jnp.ndarray) -> jnp.ndarray:
        """Order-l spherical Bessel function of the first kind with derivative.

        This implementation is based on the algorithm developed by L.-W. Cai
        and published in Computer Physics Communications 182 (2011) 664,
        doi:10.1016/j.cpc.2010.11.019 . Whereas Cai's focus is on complex
        numbers, this implementation is intended only for use with real
        arguments in JAX-based differentiable programs. Moreover, the potential
        arguments are assumed to be bounded from above by the order of the
        function.The condition r >= 0 is not assessed.

        Args:
            r: The radial coordinate.

        Returns:
            The values of the spherical Bessel function of the first
            kind and the values of its derivative up to order l, in a tuple.
        """
        plus_1 = jnp.zeros_like(r, dtype=dtype)
        temp = initial_prefactor * r * jnp.ones_like(r, dtype=dtype)

        # Iterate from the starting order to the desired one.
        for i in range(starting_order - order):
            minus_1 = (2 * (starting_order - i) + 1) * temp / r - plus_1
            plus_1, temp = temp, minus_1

        unnormalized, unnormalized_derivative = [], []
        unnormalized.append(minus_1)
        unnormalized_derivative.append(order * minus_1 / r - plus_1)

        # And then continue all the way to zero to recover the right
        # normalization.
        for i in range(order):
            minus_1 = (2 * (order - i) + 1) * temp / r - plus_1
            plus_1, temp = temp, minus_1
            unnormalized.append(minus_1)
            unnormalized_derivative.append((order-i-1) * minus_1 / r - plus_1)

        # Generally speaking, we use the explicit form of j_0 to obtain the
        # right normalization for our function, but close to the zeros of j_0
        # it is more convenient to use j_1 instead.
        order_0 = _j_0(r)
        order_1 = _j_1(r)
        prefactor = jnp.where(
            jnp.fabs(order_0) >= jnp.fabs(order_1),
            order_0 / jnp.asarray(minus_1, dtype=dtype),
            order_1 / jnp.asarray(plus_1, dtype=dtype)
        )
        normalized = prefactor * jnp.flip(jnp.asarray(unnormalized, dtype=dtype))
        normalized_derivative = (prefactor *
                                 jnp.flip(jnp.asarray(unnormalized_derivative, dtype=dtype)))

        return normalized, normalized_derivative

    @jax.custom_jvp
    @partial(jnp.vectorize,
             signature="()->(l)" if output_all else "()->()")
    def j_l(r: jnp.ndarray) -> jnp.ndarray:
        """Order-l spherical Bessel function of the first kind.

        This function just dispatches each argument to the right implementation
        based on whether r is lower than the order of the function or not. The
        condition r >= 0 is not assessed.

        Args:
            r: The radial coordinate.

        Returns:
            The values of the order-l spherical Bessel function of the first
                kind.
        """
        r = jnp.asarray(r, dtype=dtype)

        Cai_values = j_l_Cai(jnp.clip(r, a_max=order))[0]
        upward_l_values = j_l_upward(r)[0]
        condition = abs(r) < order                                              # TODO: not clear if this should be abs(r) or r.real
        selected_values = (condition * Cai_values +
                           (1. - condition) * upward_l_values)

        if not output_all:
            return selected_values[-1]
        return selected_values

    # Since the derivatives of the spherical Bessel functions are inexpensive
    # to calculate, it pays off to define a custom jvp rule for them.
    @j_l.defjvp
    def j_l_jvp(primals, tangents):
        r, = primals
        r_dot, = tangents
        Cai, Cai_dot = j_l_Cai(jnp.clip(r, a_max=order))
        upward, upward_dot = j_l_upward(r)
        condition = abs(r) < order                                              # TODO: not clear if this should be abs(r) or r.real
        primal_out = condition * Cai + (1. - condition) * upward
        tangent_out = (
            condition * Cai_dot + (1. - condition) * upward_dot
        ) * r_dot

        if not output_all:
            return primal_out[-1], tangent_out[-1]
        return primal_out, tangent_out

    return j_l


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import scipy as sp
    import scipy.special

    ORDER = 10

    returned_function = create_j_l(ORDER, dtype=onp.float32)
    r = jnp.linspace(1e-3, 50., num=1001)
    value = returned_function(r)
    reference = sp.special.spherical_jn(ORDER, r)

    plt.figure()
    plt.plot(r, value, label="value")
    plt.plot(r, reference, label="reference")
    plt.axvline(x=ORDER, color="#666666", lw=2.)
    plt.ylim(reference.min(), reference.max())
    plt.xlabel("$r$")
    plt.ylabel(f"$j_{ORDER}$")
    plt.legend(loc="best")
    plt.tight_layout()

    derivative_function = jnp.vectorize(jax.grad(returned_function))
    derivative_value = derivative_function(r)
    derivative_reference = scipy.special.spherical_jn(ORDER, r, derivative=True)

    plt.figure()
    plt.plot(r, derivative_value, label="derivative")
    plt.plot(r, derivative_reference, label="reference")
    plt.axvline(x=ORDER, color="#666666", lw=2.)
    plt.ylim(derivative_reference.min(), derivative_reference.max())
    plt.xlabel("$r$")
    plt.ylabel(f"$j'_{ORDER}$")
    plt.legend(loc="best")
    plt.tight_layout()

    plt.show()



#!/usr/bin/env python
# Copyright 2019-2022 The NeuralIL contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Code used to calculate the coefficients of the spherical Bessel functions of
# the first kind. Modelled after the the Apache-licensed reference
# implementation of the spherical Bessel descriptors
# (doi: 10.1063/1.5111045 , https://github.com/harharkh/sb_desc ).

import jax
import numpy as np

from . import functions, roots


class SphericalBesselCoefficients:
    """Calculator and cache of the coefficients of the descriptors.

    The coefficients are available directly through the attributes c_1 and c_2.

    Args:
        n_max: The maximum value of n to be used when generating the spherical
            Bessel descriptors. This affects which coefficients are calculated.
        dtype: The floating-point type to be used for the calculations.

    Raises:
        ValueError if n_max is negative.
    """
    def __init__(self, n_max: int, dtype: np.dtype = np.float32):
        if n_max < 0:
            raise ValueError("n_max cannot be negative")
        self.n_max = n_max
        # We need to go one step beyond n_max to get all the coefficients we
        # need, so we keep an internal version incremented by one.
        self.__n_max = self.n_max + 1
        # Initialize the table of Bessel roots.
        self.roots = roots.SphericalBesselRoots(self.__n_max, dtype)
        # Create the tables to store the coefficients. Just like for the
        # roots, we store them regular arrays for convenience but only use
        # half of each array. Non initialized elements are set to NaN.
        self.c_1 = (np.nan * np.ones((self.__n_max + 1, self.__n_max + 1)
                                    )).astype(dtype)
        self.c_2 = (np.nan * np.ones((self.__n_max + 1, self.__n_max + 1)
                                    )).astype(dtype)
        for order in range(self.__n_max + 1):
            function = jax.jit(functions.create_j_l(order + 1, dtype))
            u_0 = self.roots.table[order, :self.__n_max - order + 1]
            u_1 = self.roots.table[order, 1:self.__n_max - order + 2]
            coeff = np.sqrt(2. / (u_0 * u_0 + u_1 * u_1))
            self.c_1[order, :self.__n_max - order +
                     1] = u_1 / function(u_0) * coeff
            self.c_2[order, :self.__n_max - order +
                     1] = u_0 / function(u_1) * coeff


if __name__ == "__main__":
    import datetime
    import timeit

    test_n_max = 139

    start = timeit.default_timer()
    coefficients = SphericalBesselCoefficients(test_n_max)
    end = timeit.default_timer()
    print("Time elapsed:", str(datetime.timedelta(seconds=end - start)))
